/**
 * Copyright 2011 J Adam Crain (jadamcrain@gmail.com)
 *
 * Licensed to J Adam Crain under one or more contributor license agreements.
 * See the NOTICE file distributed with this work for additional information
 * regarding copyright ownership. J Adam Crain licenses this file to you under
 * the Apache License, Version 2.0 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of the License
 * at:
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */
package com.automatak.osgi4s.impl

import org.osgi.framework.{ ServiceReference, BundleContext, ServiceRegistration }
import org.osgi.util.tracker.ServiceTracker

import org.osgi.service.cm.ConfigurationAdmin
import java.util.Dictionary
import collection.JavaConversions._
import collection.immutable.Map

case class ServiceWithReference[A](ref: ServiceReference, service: A)

final class DecoratedBundleContext(ctx: BundleContext) {
  assert(ctx != null)

  /**
   * Type-safe method for publishing a service to OSGi
   * @param service Service to inject into OSGi
   * @param properties Scala map of properties
   * @param interface1 Optional interface1 to publish
   * @param interface2 Optional interface2 to publish
   * @param interface3 Optional interface3 to publish
   * @tparam S Type of the service
   * @tparam I1 Type of interface1
   * @tparam I2 Type of interface2
   * @tparam I3 Type of interface3
   * @return ServiceRegistration
   */
  def publishService[S <: AnyRef, I1 >: S <: AnyRef, I2 >: S <: AnyRef, I3 >: S <: AnyRef](
    service: S,
    properties: Map[String, String] = Map.empty,
    interface1: Option[Class[I1]] = None,
    interface2: Option[Class[I1]] = None,
    interface3: Option[Class[I1]] = None): ServiceRegistration = {

    // If no interfaces are supplied, published all interfaces for S
    val interfaces: List[String] = List(interface1, interface2, interface3).flatten.map(_.getName) match {
      case Nil => service.getClass.getInterfaces.map(_.getName).toList
      case x => x
    }

    val prop = DictionaryHelpers.convert(if (properties.isEmpty) null else properties)

    ctx.registerService(interfaces.toArray, service, prop)
  }

  def watchService[I <: AnyRef](interface: Class[I])(handler: ServiceEvent[I] => Unit): ServiceTracker = {
    val tracker = new DefaultServiceTracker[I](ctx, interface)(handler)
    tracker.open()
    tracker
  }

  /**
   * Get a reference to a particular service
   * @param interface Interface of the service to get
   * @tparam I Interface of the service to get
   * @return Some if the service is found, None otherwise
   */
  def getService[I <: AnyRef](interface: Class[I], filter: Map[String, String] = Map.empty): Option[ServiceWithReference[I]] = {

    def onService(ref: ServiceReference): Option[ServiceWithReference[I]] = {

      def serviceHasKeyValue(kv: (String, String)): Boolean = Option(ref.getProperty(kv._1)) match {
        case Some(v) => v == kv._2
        case None => false
      }

      if (filter.forall(serviceHasKeyValue)) {

        Option(ctx.getService(ref)) match {
          case Some(x) =>
            Some(ServiceWithReference(ref, x.asInstanceOf[I]))
          case None =>
            None
        }

      } else None

    }

    getServiceReference(interface).flatMap(onService)
  }

  def invokeService[I <: AnyRef, R](interface: Class[I], filter: Map[String, String] = Map.empty)(fun: I => R): Option[R] = {
    getService(interface, filter) match {
      case Some(x) =>
        try {
          Some(fun(x.service))
        } finally {
          ctx.ungetService(x.ref)
        }
      case None => None
    }
  }

  private def getServiceReference[I <: AnyRef](interface: Class[I]): Option[ServiceReference] =
    Option(ctx.getServiceReference(interface.getName))

  def loadConfigIfPresent(pid: String): Option[Map[String, String]] = {
    def convertMap(dict: Dictionary[_, _]): Map[String, String] = {
      dict.foldLeft(Map.empty[String, String])((sum, kv) => sum + (kv._1.toString -> kv._2.toString))
    }
    invokeService(classOf[ConfigurationAdmin]) { ca =>
      for { //easy way to avoid all the freaking nulls java returns
        cfg <- Option(ca.getConfiguration(pid))
        prop <- Option(cfg.getProperties)
      } yield convertMap(prop)
    }.flatten
  }

  def loadConfig(pids: List[String]): Map[String, String] = {
    pids.flatMap(loadConfigIfPresent).reduce(_ ++ _)
  }

}
